# tensorrt-lib

import os
import tensorrt as trt
import pycuda.autoinit
import pycuda.driver as cuda
from calibrator import Calibrator
from torch.autograd import Variable
import torch
import numpy as np
import time

# add verbose
TRT_LOGGER = trt.Logger(trt.Logger.VERBOSE)  # ** engine可视化 **


# create tensorrt-engine
# fixed and dynamic
def get_engine(
    max_batch_size=1,
    onnx_file_path="",
    engine_file_path="",
    fp16_mode=False,
    int8_mode=False,
    calibration_stream=None,
    calibration_table_path="",
    save_engine=False,
    dynamic=False,
):
    """Attempts to load a serialized engine if available, otherwise builds a new TensorRT engine and saves it."""

    def build_engine(max_batch_size, save_engine):
        """Takes an ONNX file and creates a TensorRT engine to run inference with"""
        with trt.Builder(TRT_LOGGER) as builder, builder.create_network(
            1
        ) as network, trt.OnnxParser(network, TRT_LOGGER) as parser:

            # parse onnx model file
            if not os.path.exists(onnx_file_path):
                quit("ONNX file {} not found".format(onnx_file_path))
            print("Loading ONNX file from path {}...".format(onnx_file_path))
            with open(onnx_file_path, "rb") as model:
                print("Beginning ONNX file parsing")
                parser.parse(model.read())
                assert (
                    network.num_layers > 0
                ), "Failed to parse ONNX model. \
                            Please check if the ONNX model is compatible "
            print("Completed parsing of ONNX file")
            print(
                "Building an engine from file {}; this may take a while...".format(
                    onnx_file_path
                )
            )

            # build trt engine
            builder.max_batch_size = max_batch_size
            if not dynamic:
                builder.max_workspace_size = 1 << 30  # 1GB
                builder.fp16_mode = fp16_mode
                if int8_mode:
                    builder.int8_mode = int8_mode
                    assert (
                        calibration_stream
                    ), "Error: a calibration_stream should be provided for int8 mode"
                    builder.int8_calibrator = Calibrator(
                        ["input"], calibration_stream, calibration_table_path
                    )
                    print("Int8 mode enabled")
                engine = builder.build_cuda_engine(network)
            else:
                config = builder.create_builder_config()
                config.max_workspace_size = 1 << 30  # 1GB
                profile = builder.create_optimization_profile()
                profile.set_shape(
                    network.get_input(0).name,
                    (1, 3, 200, 200),
                    (1, 3, 608, 448),
                    (1, 3, 1200, 1200),
                )
                # dynamic_engine fp16 set
                if fp16_mode:
                    config.set_flag(trt.BuilderFlag.FP16)
                # dynamic_engine int8 set
                if int8_mode:
                    config.set_flag(trt.BuilderFlag.INT8)
                    assert (
                        calibration_stream
                    ), "Error: a calibration_stream should be provided for int8 mode"
                    # choose an calibration profile
                    # config.set_calibration_profile(profile)
                    config.int8_calibrator = Calibrator(
                        ["input"], calibration_stream, calibration_table_path
                    )
                    print("Int8 mode enabled")
                # choose an optimization profile
                config.add_optimization_profile(profile)
                engine = builder.build_engine(network, config)
            # If errors happend when executing builder.build_cuda_engine(network),
            # a None-Type object would be returned
            if engine is None:
                print("Failed to create the engine")
                return None
            print("Completed creating the engine")
            if save_engine:
                with open(engine_file_path, "wb") as f:
                    f.write(engine.serialize())
            return engine

    if os.path.exists(engine_file_path):
        # If a serialized engine exists, load it instead of building a new one.
        print("Reading engine from file {}".format(engine_file_path))
        with open(engine_file_path, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
            return runtime.deserialize_cuda_engine(f.read())
    else:
        return build_engine(max_batch_size, save_engine)


class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        """Within this context, host_mom means the cpu memory and device means the GPU memory"""
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


# fixed_engine buffer
def allocate_buffers(engine):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    for binding in engine:
        size = trt.volume(engine.get_binding_shape(binding)) * engine.max_batch_size
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
    return inputs, outputs, bindings, stream


# dynamic_engine buffer
def allocate_buffers_v2(engine, h_, w_):
    inputs = []
    outputs = []
    bindings = []
    stream = cuda.Stream()
    print("engine.get_binding_format_desc", engine.get_binding_format_desc(0))
    for count, binding in enumerate(engine):
        print("binding:", binding)
        size = (
            trt.volume(engine.get_binding_shape(binding))
            * engine.max_batch_size
            * h_
            * w_
        )
        dtype = trt.nptype(engine.get_binding_dtype(binding))
        print("dtype:", dtype)
        # Allocate host and device buffers
        host_mem = cuda.pagelocked_empty(size, dtype)
        device_mem = cuda.mem_alloc(host_mem.nbytes)
        # Append the device buffer to device bindings.
        bindings.append(int(device_mem))
        # Append to the appropriate list.
        if engine.binding_is_input(binding):
            inputs.append(HostDeviceMem(host_mem, device_mem))
        else:
            outputs.append(HostDeviceMem(host_mem, device_mem))
        print("size:", size)
        print("input:", inputs)
        print("output:", outputs)
        print("------------------")
    return inputs, outputs, bindings, stream


# fixed_engine inference
def do_inference(context, bindings, inputs, outputs, stream, batch_size=1):
    # Transfer data from CPU to GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # GPU Run inference.
    context.execute_async(
        batch_size=batch_size, bindings=bindings, stream_handle=stream.handle
    )
    # context.execute(batch_size=batch_size, bindings=bindings)
    # Transfer predictions from GPU to CPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


# dynamic_engine inference
def do_inference_v2(context, bindings, inputs, outputs, stream, h_, w_, binding_id):
    # set the input dimensions
    context.set_binding_shape(binding_id, (1, 3, h_, w_))
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    infer_start = time.time()
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
    # context.execute_v2(bindings=bindings)
    infer_end = time.time()
    infer_time = infer_end - infer_start
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs], infer_time


# numpy_array reshape
def postprocess_the_outputs(h_outputs, shape_of_output):
    h_outputs = h_outputs.reshape(*shape_of_output).copy()
    return h_outputs


# tensor to numpy
def to_numpy(tensor):
    return (
        tensor.detach().cpu().numpy() if tensor.requires_grad else tensor.cpu().numpy()
    )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

sys.path.append("../../../..")
import os
import math
import argparse
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.autograd import Variable
from torch.nn import init
from models import nin_gc, nin, resnet

import quantize


def setup_seed(seed):
    torch.manual_seed(seed)
    # torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = True


def save_state(model, best_acc):
    print("==> Saving model ...")
    state = {
        "best_acc": best_acc,
        "state_dict": model.state_dict(),
    }
    state_copy = state["state_dict"].copy()
    for key in state_copy.keys():
        if "module" in key:
            state["state_dict"][key.replace("module.", "")] = state["state_dict"].pop(
                key
            )
    if args.model_type == 0:
        if args.bn_fuse:
            if args.prune_quant or args.prune_qaft:
                torch.save(
                    {
                        "cfg": cfg,
                        "best_acc": best_acc,
                        "state_dict": state["state_dict"],
                    },
                    "models_save/nin_bn_fused.pth",
                )
            else:
                torch.save(state, "models_save/nin_bn_fused.pth")
        else:
            if args.prune_quant or args.prune_qaft:
                torch.save(
                    {
                        "cfg": cfg,
                        "best_acc": best_acc,
                        "state_dict": state["state_dict"],
                    },
                    "models_save/nin.pth",
                )
            else:
                torch.save(state, "models_save/nin.pth")
    elif args.model_type == 1:
        if args.bn_fuse:
            if args.prune_quant or args.prune_qaft:
                torch.save(
                    {
                        "cfg": cfg,
                        "best_acc": best_acc,
                        "state_dict": state["state_dict"],
                    },
                    "models_save/nin_gc_bn_fused.pth",
                )
            else:
                torch.save(state, "models_save/nin_gc_bn_fused.pth")
        else:
            if args.prune_quant or args.prune_qaft:
                torch.save(
                    {
                        "cfg": cfg,
                        "best_acc": best_acc,
                        "state_dict": state["state_dict"],
                    },
                    "models_save/nin_gc.pth",
                )
            else:
                torch.save(state, "models_save/nin_gc.pth")
    else:
        if args.bn_fuse:
            torch.save(state, "models_save/resnet_bn_fused.pth")
        else:
            torch.save(state, "models_save/resnet.pth")


def adjust_learning_rate(optimizer, epoch):
    update_list = [80, 130, 180, 230, 280]
    if epoch in update_list:
        for param_group in optimizer.param_groups:
            param_group["lr"] = param_group["lr"] * 0.1
    return


def train(epoch):
    model.train()

    batch_num = 0
    for batch_idx, (data, target) in enumerate(trainloader):
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        loss = criterion(output, target)

        # PTQ doesn't need backward
        if not args.ptq_control:
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            if batch_idx % 100 == 0:
                print(
                    "Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}\tLR: {}".format(
                        epoch,
                        batch_idx * len(data),
                        len(trainloader.dataset),
                        100.0 * batch_idx / len(trainloader),
                        loss.data.item(),
                        optimizer.param_groups[0]["lr"],
                    )
                )
        else:
            batch_num += 1
            if batch_num > args.ptq_batch:
                break
            print("Batch:", batch_num)
    return


def test():
    global best_acc
    model.eval()
    test_loss = 0
    correct = 0

    for data, target in testloader:
        if not args.cpu:
            data, target = data.cuda(), target.cuda()
        data, target = Variable(data), Variable(target)
        output = model(data)
        test_loss += criterion(output, target).data.item()
        pred = output.data.max(1, keepdim=True)[1]
        correct += pred.eq(target.data.view_as(pred)).cpu().sum()
    acc = 100.0 * float(correct) / len(testloader.dataset)

    if acc > best_acc:
        best_acc = acc
        save_state(model, best_acc)
    average_test_loss = test_loss / (len(testloader.dataset) / args.eval_batch_size)

    print(
        "\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.2f}%)".format(
            average_test_loss,
            correct,
            len(testloader.dataset),
            100.0 * float(correct) / len(testloader.dataset),
        )
    )

    print("Best Accuracy: {:.2f}%\n".format(best_acc))
    return


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--cpu", action="store_true", help="set if only CPU is available"
    )
    parser.add_argument("--gpu_id", action="store", default="", help="gpu_id")
    parser.add_argument(
        "--data", action="store", default="../../../../data", help="dataset path"
    )
    parser.add_argument(
        "--lr", action="store", default=0.01, help="the intial learning rate"
    )
    parser.add_argument(
        "--wd", action="store", default=1e-5, help="the intial learning rate"
    )
    # prune_quant
    parser.add_argument(
        "--prune_quant",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the prune_quant model",
    )
    # refine
    parser.add_argument(
        "--refine",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the float_refine model",
    )
    # resume
    parser.add_argument(
        "--resume",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the resume model",
    )
    parser.add_argument("--batch_size", type=int, default=32)
    parser.add_argument("--eval_batch_size", type=int, default=32)
    parser.add_argument("--num_workers", type=int, default=2)
    parser.add_argument(
        "--start_epochs",
        type=int,
        default=1,
        metavar="N",
        help="number of epochs to train_start",
    )
    parser.add_argument(
        "--end_epochs",
        type=int,
        default=300,
        metavar="N",
        help="number of epochs to train_end",
    )
    # W/A — bits
    parser.add_argument("--w_bits", type=int, default=8)
    parser.add_argument("--a_bits", type=int, default=8)
    # bn融合标志位
    parser.add_argument(
        "--bn_fuse", action="store_true", help="batch-normalization fuse"
    )
    # bn融合校准标志位
    parser.add_argument(
        "--bn_fuse_calib",
        action="store_true",
        help="batch-normalization fuse calibration",
    )
    # 量化方法选择
    parser.add_argument(
        "--q_type", type=int, default=0, help="quant_type:0-symmetric, 1-asymmetric"
    )
    # 量化级别选择
    parser.add_argument(
        "--q_level", type=int, default=0, help="quant_level:0-per_channel, 1-per_layer"
    )
    # weight_observer选择
    parser.add_argument(
        "--weight_observer",
        type=int,
        default=0,
        help="quant_weight_observer:0-MinMaxObserver, 1-MovingAverageMinMaxObserver",
    )
    # pretrained_model标志位
    parser.add_argument(
        "--pretrained_model", action="store_true", help="pretrained_model"
    )
    # qaft标志位
    parser.add_argument(
        "--qaft", action="store_true", help="quantization-aware-finetune"
    )
    # prune_qaft
    parser.add_argument(
        "--prune_qaft",
        default="",
        type=str,
        metavar="PATH",
        help="the path to the prune_qaft model",
    )
    # ptq_observer
    parser.add_argument("--ptq", action="store_true", help="post-training-quantization")
    # ptq_control
    parser.add_argument("--ptq_control", action="store_true", help="ptq control flag")
    # ptq_percentile
    parser.add_argument(
        "--percentile", type=float, default=0.999999, help="the percentile of ptq"
    )
    # ptq_batch
    parser.add_argument("--ptq_batch", type=int, default=200, help="the batch of ptq")
    parser.add_argument(
        "--model_type", type=int, default=1, help="model type:0-nin,1-nin_gc,2-resnet"
    )
    args = parser.parse_args()
    print("==> Options:", args)

    if args.gpu_id:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu_id

    setup_seed(1)

    print("==> Preparing data..")
    transform_train = transforms.Compose(
        [
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )
    transform_test = transforms.Compose(
        [
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
        ]
    )

    trainset = torchvision.datasets.CIFAR10(
        root=args.data, train=True, download=True, transform=transform_train
    )
    trainloader = torch.utils.data.DataLoader(
        trainset, batch_size=args.batch_size, shuffle=True, num_workers=args.num_workers
    )

    testset = torchvision.datasets.CIFAR10(
        root=args.data, train=False, download=True, transform=transform_test
    )
    testloader = torch.utils.data.DataLoader(
        testset,
        batch_size=args.eval_batch_size,
        shuffle=False,
        num_workers=args.num_workers,
    )

    classes = (
        "plane",
        "car",
        "bird",
        "cat",
        "deer",
        "dog",
        "frog",
        "horse",
        "ship",
        "truck",
    )

    if args.prune_quant:
        print("******Prune Quant model******")
        # checkpoint = torch.load('../prune/models_save/nin_refine.pth')
        checkpoint = torch.load(args.prune_quant)
        cfg = checkpoint["cfg"]
        if args.model_type == 0:
            model = nin.Net(cfg=checkpoint["cfg"])
        else:
            model = nin_gc.Net(cfg=checkpoint["cfg"])
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = 0
        print("***ori_model***\n", model)
        quantize.prepare(
            model,
            inplace=True,
            a_bits=args.a_bits,
            w_bits=args.w_bits,
            q_type=args.q_type,
            q_level=args.q_level,
            weight_observer=args.weight_observer,
            bn_fuse=args.bn_fuse,
            bn_fuse_calib=args.bn_fuse_calib,
            pretrained_model=args.pretrained_model,
            qaft=args.qaft,
            ptq=args.ptq,
            percentile=args.percentile,
        )
        print("\n***quant_model***\n", model)
    elif args.prune_qaft:
        print("******Prune QAFT model******")
        # checkpoint = torch.load('models_save/nin_bn_fused.pth')
        checkpoint = torch.load(args.prune_qaft)
        cfg = checkpoint["cfg"]
        if args.model_type == 0:
            model = nin.Net(cfg=checkpoint["cfg"])
        else:
            model = nin_gc.Net(cfg=checkpoint["cfg"])
        print("***ori_model***\n", model)
        quantize.prepare(
            model,
            inplace=True,
            a_bits=args.a_bits,
            w_bits=args.w_bits,
            q_type=args.q_type,
            q_level=args.q_level,
            weight_observer=args.weight_observer,
            bn_fuse=args.bn_fuse,
            bn_fuse_calib=args.bn_fuse_calib,
            pretrained_model=args.pretrained_model,
            qaft=args.qaft,
            ptq=args.ptq,
            percentile=args.percentile,
        )
        print("\n***quant_model***\n", model)
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = checkpoint["best_acc"]
    elif args.refine:
        print("******Float Refine model******")
        # checkpoint = torch.load('models_save/nin.pth')
        checkpoint = torch.load(args.refine)
        if args.model_type == 0:
            model = nin.Net()
        elif args.model_type == 1:
            model = nin_gc.Net()
        else:
            model = resnet.resnet18()
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = 0
        print("***ori_model***\n", model)
        quantize.prepare(
            model,
            inplace=True,
            a_bits=args.a_bits,
            w_bits=args.w_bits,
            q_type=args.q_type,
            q_level=args.q_level,
            weight_observer=args.weight_observer,
            bn_fuse=args.bn_fuse,
            bn_fuse_calib=args.bn_fuse_calib,
            pretrained_model=args.pretrained_model,
            qaft=args.qaft,
            ptq=args.ptq,
            percentile=args.percentile,
        )
        print("\n***quant_model***\n", model)
    elif args.resume:
        print("******Reume model******")
        # checkpoint = torch.load('models_save/nin.pth')
        checkpoint = torch.load(args.resume)
        if args.model_type == 0:
            model = nin.Net()
        elif args.model_type == 1:
            model = nin_gc.Net()
        else:
            model = resnet.resnet18()
        print("***ori_model***\n", model)
        quantize.prepare(
            model,
            inplace=True,
            a_bits=args.a_bits,
            w_bits=args.w_bits,
            q_type=args.q_type,
            q_level=args.q_level,
            weight_observer=args.weight_observer,
            bn_fuse=args.bn_fuse,
            bn_fuse_calib=args.bn_fuse_calib,
            pretrained_model=args.pretrained_model,
            qaft=args.qaft,
            ptq=args.ptq,
            percentile=args.percentile,
        )
        print("\n***quant_model***\n", model)
        model.load_state_dict(checkpoint["state_dict"])
        best_acc = checkpoint["best_acc"]
    else:
        print("******Initializing model******")
        if args.model_type == 0:
            model = nin.Net()
        elif args.model_type == 1:
            model = nin_gc.Net()
        else:
            model = resnet.resnet18()
        best_acc = 0
        for m in model.modules():
            if isinstance(m, nn.Conv2d):
                init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                init.normal_(m.weight, 0, 0.01)
                if m.bias is not None:
                    init.zeros_(m.bias)
        print("***ori_model***\n", model)
        quantize.prepare(
            model,
            inplace=True,
            a_bits=args.a_bits,
            w_bits=args.w_bits,
            q_type=args.q_type,
            q_level=args.q_level,
            weight_observer=args.weight_observer,
            bn_fuse=args.bn_fuse,
            bn_fuse_calib=args.bn_fuse_calib,
            pretrained_model=args.pretrained_model,
            qaft=args.qaft,
            ptq=args.ptq,
            percentile=args.percentile,
        )
        print("\n***quant_model***\n", model)

    if not args.cpu:
        model.cuda()
        model = torch.nn.DataParallel(
            model, device_ids=range(torch.cuda.device_count())
        )

    base_lr = float(args.lr)
    param_dict = dict(model.named_parameters())
    params = []
    for key, value in param_dict.items():
        params += [{"params": [value], "lr": base_lr, "weight_decay": args.wd}]

    criterion = nn.CrossEntropyLoss()
    optimizer = optim.Adam(params, lr=base_lr, weight_decay=args.wd)

    if args.ptq_control:
        args.end_epochs = 2
        print("ptq is doing...")
    for epoch in range(args.start_epochs, args.end_epochs):
        adjust_learning_rate(optimizer, epoch)
        train(epoch)
        test()
    if args.ptq_control:
        print("ptq is done")
